/*****************************************************************************
 * Copyright 2013 celum Slovakia s r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/

package com.celum.dbtool.configuration;

import com.celum.dbtool.Db;
import com.celum.dbtool.PatchStrategy;

/**
 * Simplify configuration of Db where is used patching strategy
 * and you have to determine just table where will be patches stored.
 *
 * example of configuration what is ready for patching.
 * <code>
 *     Db db = Db.withJdbcConnection(connection).createDefaultDbConfiguration("PATCHES").forPatching().andDbScriptsInPackage("com.celum.db");
 *     db.patch();
 * </code>

 *
 * example of configuration with self-initialization of PATCHES table:
 * <code>
 *     Db db = Db.withJdbcConnection(connection).createDefaultDbConfiguration("PATCHES").withSelfInit().andDbScriptsInPackage("com.celum.db");
 *     db.patch();
 * </code>
 */
public class DefaultMigrationConfiguration {

    /** general configuration for DB Engine */
    private final Db.DbBuilder builder;

    /**
     * Constructor
     */
    public DefaultMigrationConfiguration(Db.DbBuilder builder, String patchesTable)
    {
        this.builder = builder;
        builder.patchAs(PatchStrategy.CONDITIONAL);
        builder.withVariable("patches", patchesTable);
        builder.withVersionSql("SELECT COUNT (*) FROM $patches WHERE version LIKE '$version'");
        builder.withVersionUpdate("INSERT INTO $patches (version, name) VALUES ('$version', '$name')");
    }


    /**
     * setup Db for performing 'patches' but before patches are applied, the
     * method initialize DB (means create PATCHES table). If the initialization failed,
     * it means probably the DB is initialized and the method continue.
     */
    public Db.DbBuilder withSelfInit()
    {
        try {
            builder.andDbScriptsInPackage("com.celum.dbtool.configuration.init").run();
        } catch (Exception e) {
            //do nothing
        }
        return builder;
    }


    /**
     * setup DB for performing 'install()' and 'patch()' methods. Before you
     * call these methods, you have to determine the scripts source first.
     */
    public Db.DbBuilder forPatching()
    {
        return builder;
    }


    /**
     * do self-init when IF condition is true. If it's false
     * the 'forPatching' is used
     */
    public Db.DbBuilder withSelfInitIf(boolean condition)
    {
        if (condition) {
            return withSelfInit();
        } else {
            return forPatching();
        }
    }

}
