/*****************************************************************************
 * Copyright 2012 celum Slovakia s r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package com.celum.dbtool.resource;

import com.celum.dbtool.DbException;
import com.celum.dbtool.step.*;

import java.io.File;
import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.sql.Connection;
import java.util.*;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/**
 * Script resource take given package where are steps
 * placed.
 *
 * @author Zdenko Vrabel (zdenko.vrabel@celum.com)
 */
public class PackageResource extends DbStepResource
{
    private final static String JAR_EXTENSION = ".jar!";

    /** package where are steps stored */
    private final String pckgName;

    /** classloader for step loading, default is system classloader */
    private final ClassLoader loader;


    /**
     * Constructor
     * @param pckg package instance
     */
    public PackageResource(Package pckg)
    {
        this(pckg.getName());
    }


    /**
     * Constructor
     * @param pckgName package name with '.'
     */
    public PackageResource(String pckgName)
    {
        this(pckgName, ClassLoader.getSystemClassLoader());
    }


    /**
     * Constructor
     * @param pckgName
     * @param loader
     */
    public PackageResource(String pckgName, ClassLoader loader)
    {
        this.pckgName = pckgName;
        this.loader = loader;
    }


    /**
     * {@inheritDoc}
     */
    @Override
    public List<DbStep> getSteps()
    {
        try {
            String relPath = pckgName.replace('.', '/');
            List<DbStep> steps = new LinkedList<DbStep>();
            Enumeration<URL> resources = loader.getResources(relPath);

            //go through all resources
            while (resources.hasMoreElements()) {
                URL resource = resources.nextElement();
                if (resource.getProtocol().compareToIgnoreCase("file") == 0) {
                    steps = loadFromDir(resource, steps);
                }
                else if (resource.getProtocol().compareToIgnoreCase("jar") == 0) {
                    steps = loadFromJar(resource, steps);
                }
            }

            return steps;
        } catch (Exception e) {
            throw new DbException("error loading steps from package " + pckgName, e);
        }
    }


    /**
     * Method scan directory which represents the JAVA package and load
     * all scripts into 'toList' list.
     */
    private List<DbStep> loadFromDir(URL pckgResource, List<DbStep> toList) throws URISyntaxException
    {
        File pckgDir = new File(pckgResource.toURI());
        File[] files = pckgDir.listFiles();
        for (File f : files) {
            if (f.isFile()) {
                StepType type = StepType.getTypeOfFile(f.getName());

                switch (type) {
                    case UNSUPPORTED:
                        break;

                    case JAVA:
                        String className = f.getName().substring(0, f.getName().length() - ".class".length() );
                        loadClass(pckgName + "." + className, toList);
                        break;

                    default:
                        String name = pckgName.replace('.', '/') + "/" + f.getName();
                        Version v = Version.of(f.getName());
                        toList.add(new ClasspathStep(name, v, type, name, loader));
                        break;
                }
            }
        }

        return toList;
    }


    /**
     * Method scan package inside JAR and load scripts into 'toList'
     */
    private List<DbStep> loadFromJar(URL pckgResource, List<DbStep> toList) throws URISyntaxException, IOException
    {
        String name = pckgName.replace('.', '/');

        //get and open JAR file
        String resourceName = pckgResource.toString();
        int pos = resourceName.indexOf(JAR_EXTENSION);
        String jarPath = resourceName.substring(4, pos + JAR_EXTENSION.length() - 1);
        JarFile jarFile = new JarFile(new File(new URI(jarPath)));

        //go through JAR entries
        Enumeration<JarEntry> entries = jarFile.entries();
        while (entries.hasMoreElements()) {
            JarEntry entry = entries.nextElement();
            if (!entry.isDirectory()) {
                String entryName = entry.getName();
                if (entryName.startsWith(name)) {
                    //create SQL step
                    StepType type = StepType.getTypeOfFile(entryName);
                    switch (type) {
                        case UNSUPPORTED:
                            break;

                        case JAVA:
                            entryName = entryName.replace('/', '.');
                            entryName = entryName.substring(0, entryName.length() - ".class".length());
                            loadClass(entryName, toList);
                            break;

                        default:
                            Version v = Version.of(entryName.substring(name.length()));
                            toList.add(new ClasspathStep(entryName, v, type, entryName, loader));
                            break;

                    }
                }
            }
        }

        return toList;
    }


    /**
     * load class with static method 'dbmain' as step.
     */
    private List<DbStep> loadClass(String className, List<DbStep> toList)
    {
        try {
            Class<?> cls = loader.loadClass(className);

            Version v = getStepVersionFromClass(cls);
            if (v != null) {
                Method m = cls.getMethod("dbmain", Connection.class, Map.class);
                toList.add(new JavaStep(m, className, v));
            }

        } catch (ClassNotFoundException e) {
            throw new DbException("class not found:" + className);
        } catch (NoSuchMethodException e) {
            //if 'dbmain' is missing, the class is ignored
        }
        return toList;
    }


    /**
     * returns step version from method which is annotated with StepVersion.
     *
     * Replaced by getStepVersionFromClass
     */
    @Deprecated
    private Version getStepVersionFromMethod(Method m)
    {
        Annotation[] annotations = m.getDeclaredAnnotations();
        for (Annotation a : annotations) {
            if (a instanceof StepVersion) {
                StepVersion se = (StepVersion)a;
                return Version.of(se.value());
            }
        }
        return Version.NULL;
    }


    /**
     * returns step version from class what is annotated with StepVersion
     */
    private Version getStepVersionFromClass(Class<?> cls)
    {
        Annotation[] annotations = cls.getDeclaredAnnotations();
        for (Annotation a : annotations) {
            if (a instanceof StepVersion) {
                StepVersion se = (StepVersion)a;
                return Version.of(se.value());
            }
        }
        return null;
    }
}
