/*****************************************************************************
 * Copyright 2013 celum Slovakia s r.o.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *****************************************************************************/
package com.celum.dbtool.sql;

import com.celum.dbtool.DbException;
import org.apache.commons.dbutils.DbUtils;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

/**
 * Implementation executes the SQL query as statement and
 * provide you result-set manipulation like row mapping etc.
 */
public class SqlResultSetStrategy implements SqlQueryStrategy
{
    /** when query is executed, then is filled by result set */
    protected ResultSet rs;

    /** statement which is used for SQL */
    private Statement stmt;

    /** connection for JDBC */
    private final Connection dbConnection;

    /**
     * Constructor
     * @param dbConnection
     */
    public SqlResultSetStrategy(Connection dbConnection)
    {
        this.dbConnection = dbConnection;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void query(String sql)
    {
        try {
            if (sql != null && !sql.isEmpty() ) {
                stmt = dbConnection.createStatement();
                rs = stmt.executeQuery(sql);
            }
        } catch (SQLException e) {
            throw new SqlCommandException(sql, e);
        } finally {
        }
    }


    /**
     * Goes through result set and call RowMapper::map for
     * each row.
     */
    public void withRowMapping(RowMapper rm)
    {
        try {
            long rownum = 0;
            while (rs.next()) {
                rm.map(rs,rownum);
                rownum++;
            }
        } catch (SQLException e) {
            throw new DbException("SQL error:" + e.getMessage(), e);
        } finally {
            DbUtils.closeQuietly(rs);
            DbUtils.closeQuietly(stmt);
        }
    }


    /**
     * returns string value from first columnd from first row.
     * If select returns empty rows, the 'null' value is returned.
     */
    public String andReturnString()
    {
        return andReturnString(null);
    }


    /**
     * Return string value in first row, in first column. If
     * there is no row or column the defaultValue is returned.
     */
    public String andReturnString(String defaultValue)
    {
        try {
            if (rs.next()) {
                return rs.getString(1);
            } else {
                return defaultValue;
            }
        } catch (SQLException e) {
            throw new DbException("SQL error:" + e.getMessage(), e);
        } finally {
            DbUtils.closeQuietly(rs);
            DbUtils.closeQuietly(stmt);
        }
    }


    /**
     * returns long value from first row in first column. If
     * the select returns empty rows, the '0' is returned.
     */
    public long andReturnLong()
    {
        return andReturnLong(0);
    }


    /**
     * returns long value in first row, in first column. It
     * there is no row or column, the defaultValue is returned.
     */
    public long andReturnLong(long defaultValue)
    {
        try {
            if (rs.next()) {
                return rs.getLong(1);
            } else {
                return defaultValue;
            }
        } catch (SQLException e) {
            throw new DbException("SQL error:" + e.getMessage(), e);
        } finally {
            DbUtils.closeQuietly(rs);
            DbUtils.closeQuietly(stmt);
        }
    }

    /**
     * returns boolean value from first row in first column. If
     * the select returns empty rows, the 'false' is returned.
     */
    public boolean andReturnBoolean()
    {
        return andReturnBoolean(false);
    }


    /**
     * returns boolean value in first row, in first column. It
     * there is no row or column, the defaultValue is returned
     */
    public boolean andReturnBoolean(boolean defaultValue)
    {
        try {
            if (rs.next()) {
                return rs.getBoolean(1);
            } else {
                return defaultValue;
            }
        } catch (SQLException e) {
            throw new DbException("SQL error:" + e.getMessage(), e);
        } finally {
            DbUtils.closeQuietly(rs);
            DbUtils.closeQuietly(stmt);
        }
    }
}
